// use sha256;
use anyhow::{Context, Result, bail};
use std::fs;
use std::fs::remove_file;
use std::io::BufRead;
use std::io::BufReader;
use std::io::Write;
use std::path::PathBuf;
use std::time::{Duration, SystemTime};
use url::Url;

#[deny(missing_docs)]
// #[warn(missing_docs)]

/// Convert relate to absolute
#[must_use]
pub fn relative_to_absolute(base_url: &str, relative_href: &str) -> Result<String> {
    let base_url = Url::parse(base_url).context(format!("Url::parse({})", base_url))?;
    let new_url = base_url
        .join(relative_href)
        .context(format!("join({})", relative_href))?;
    Ok(new_url.to_string())
}

/*
Or maybe I should just use the replacements only, so I have some
idea where the file came from?
 */

/// Extract filename from the end of a URL.
///
/// If this doesn't have a usable path, convert url:
/// * Remove https:// part
/// * Replace / with -
#[must_use]
pub fn filename_from_url(url: &str) -> Result<String> {
    let (_, filename) = url.rsplit_once('/').context("Failed to split URL.")?;

    if filename.is_empty() {
        // Getting the filename part failed.
        // Like in cases where the url is https://go.dev/dl/
        // Which becomes go.dev-dl
        let mut path = url.to_string();
        path = path.replace("https://", "");
        path = path.replace("http://", "");
        path = path.replace("/", "-");
        if path.ends_with("-") {
            path.pop();
        }
        return Ok(path);
    }
    Ok(filename.to_string())
}

// Display a number as K, M, G, T, etc.
// pub fn display_bytes(size: u64) -> String { }

/// Save reqwest::header::HeaderMap to file.
///
/// This also stores the url in the file, so I know what URL was called for
/// this reqwest.
pub fn save_headermap(
    filename: &str,
    url: &str,
    header: &reqwest::header::HeaderMap,
) -> Result<()> {
    let mut fp = std::fs::File::create(filename)?;

    fp.write_all(format!("url: {}\n", url).as_bytes())?;
    for (key, value) in header.iter() {
        if let Ok(value) = value.to_str() {
            fp.write_all(format!("{}: {}\n", key, value).as_bytes())?;
        }
    }
    Ok(())
}

/// Load reqwest::header::HeaderMap from file.
///
/// This will have the url of the original call in the "url" section.
pub fn load_headermap(filename: &str) -> Result<reqwest::header::HeaderMap> {
    let fp = std::fs::File::open(filename)?;
    let mut buffer = BufReader::new(fp);
    let mut line = String::new();
    let mut header = reqwest::header::HeaderMap::new();

    loop {
        if buffer.read_line(&mut line).unwrap() == 0 {
            break;
        };
        let temp = line.trim_end();

        if let Some(parts) = temp.split_once(": ") {
            let head = reqwest::header::HeaderName::from_bytes(parts.0.as_bytes()).unwrap();
            if let Ok(value) = reqwest::header::HeaderValue::from_str(&parts.1) {
                header.insert(head, value);
            }
        }

        line.clear();
    }

    Ok(header)
}

/// Caching web calls
///
/// Set the directory, and we're ready to make cached web calls.
/// Since we're not storing the file in memory now, max_size isn't
/// the concern it once was.
pub struct Cache {
    /// Directory where cache is stored
    pub directory: PathBuf,
    // This is where we select async or blocking.
    /// Reqwest Client
    pub client: reqwest::blocking::Client,
    /// Vector of content-types to accept (empty=all)
    pub accept: Vec<String>,
    /// Max size of content to download (default unlimited)
    pub max_size: Option<u64>,
}

/// Status of fetch
#[allow(dead_code)]
#[derive(Debug)]
#[repr(u8)]
pub enum Status {
    /// File was downloaded.
    Fetched(std::fs::File) = 0,
    /// File was retrieved from cache.
    Cached(std::fs::File),
    /// Reqwest error (unable to connect)
    Error(reqwest::Error),
    /// Content-Type wasn't allowed, see Cache.accept.
    Unacceptable(String), // Content-Type
    /// Content was too big, see Cache.max_size.
    TooBig(u64),
    /// HTTP Error/status code.
    ErrorStatus(u16),
}

/*
Some possible content-type values:  We're only interested in a few of these...

text/css
text/javascript
text/plain
image/jpeg
image/png
iamge/gif
application/xml
application/javascript
 */

// If nothing is given for useragent, we default to the application name and version.
static APP_USER_AGENT: &str = concat!(env!("CARGO_PKG_NAME"), "/", env!("CARGO_PKG_VERSION"),);

impl Cache {
    /// Construct Cache using given directory for caching, and useragent.
    pub fn new(dir: PathBuf, useragent: Option<&str>) -> Result<Self> {
        // Verify the directory exists

        let path = dir.as_path();
        if path.exists() {
            if !path.is_dir() {
                // It exists, but it isn't a directory!  What?!
                bail!(
                    "Can't create Cache dir {}, it already exists.",
                    dir.display()
                );
            }
        } else {
            fs::create_dir_all(path).context(format!("Cache dir {}", path.display()))?;
        }

        let user_agent = if let Some(ua) = useragent {
            ua
        } else {
            APP_USER_AGENT
        };

        // This is where we select async or blocking.
        let client = reqwest::blocking::Client::builder()
            .user_agent(user_agent)
            .build()?;

        Ok(Self {
            directory: dir,
            client: client,
            accept: vec![], // Accept all content-type.
            max_size: None, // Accept any sized content.
                            // Some(256 * 1024 * 1024), // 256 MB
        })
    }

    #[allow(dead_code)]
    pub fn add_content_type(&mut self, content_type: String) {
        self.accept.push(content_type);
    }

    #[allow(dead_code)]
    pub fn set_max_size(&mut self, size: u64) {
        self.max_size = Some(size);
    }

    /// Create safe filename from url for header/content files.
    pub fn url_to_basename(url: &str) -> Result<String> {
        let filename = if url.ends_with("/") {
            ""
        } else {
            if let Some(has_file) = url.rsplit_once("/") {
                has_file.1
            } else {
                ""
            }
        };

        if filename.is_empty() {
            // Getting the filename part failed.
            // Like in cases where the url is https://go.dev/dl/
            // Which becomes go.dev-dl
            let mut path = url.to_string();
            path = path.replace("https://", "");
            path = path.replace("http://", "");
            path = path.replace("/", "-");
            path = path.replace(".", "-");
            if path.ends_with("-") {
                path.pop();
            }
            return Ok(path);
        }
        Ok(filename.to_string())
    }

    /// Expire files in the cache older then given age
    ///
    /// Use DirEntry.modified, since it updates when a file is freshened/downloaded.
    /// DirEntry.created isn't updated when file is rewritten.
    pub fn expire(&self, age: Duration) -> Result<bool> {
        let now = SystemTime::now();
        let mut result: bool = false;

        for file in std::fs::read_dir(self.directory.as_path())? {
            let file = file?;
            if let Ok(d) = file.metadata() {
                if d.is_file() {
                    // Created isn't updated if the file is fetched.  Use modified, that updates on fetch.

                    let filename = String::from(file.file_name().to_str().unwrap());

                    if filename.ends_with(".header") {
                        if let Ok(modify) = d.modified() {
                            if let Ok(delta) = now.duration_since(modify) {
                                // println!("expire {} = modified {}", filename, delta.as_secs());
                                if delta > age {
                                    // println!("Would delete: {} (and .content)", filename);
                                    let mut filepath = self.directory.join(filename);
                                    let r = remove_file(&filepath);
                                    if let Err(e) = r {
                                        println!("RemoveFile {:?}: {}", filepath, e);
                                    }
                                    // Also delete .content !
                                    filepath.set_extension("content");
                                    let r = remove_file(&filepath);
                                    if let Err(e) = r {
                                        println!("RemoveFile {:?}: {}", filepath, e);
                                    }

                                    result = true;
                                }
                            }
                        }
                    }

                    /*
                    if let Ok(access) = d.accessed() {
                        if let Ok(delta) = now.duration_since(access) {
                            println!("accessed {:?} = accessed {}", file.file_name(), delta.as_secs());
                            if delta > age {
                                println!("Expire: {:?}", file.file_name());
                            }
                        }
                    }

                    if let Ok(created) = d.created() {
                        if let Ok(delta) = now.duration_since(created) {
                            println!("expire {:?} = created {}", file.file_name(), delta.as_secs());
                            if delta > age {
                                println!("Would delete: {:?}", file.file_name());
                                result = true;
                            }
                        }
                    }
                    */
                }
            }
        }
        Ok(result)
    }

    /// Given a url, return an open file
    ///
    /// Reading from cache.
    #[allow(dead_code)]
    pub fn file(&self, url: &str) -> Option<std::fs::File> {
        let base = self
            .directory
            .as_path()
            .join(Self::url_to_basename(url).unwrap());
        if base.exists() {
            return Some(std::fs::File::open(base).unwrap());
        }
        None
    }

    // I'm not sure about using Result<Status> here...
    // It would allow for ? usage.

    /// Fetch the URL from the web
    ///
    /// This returns Status, which could be Fetched or Cached copy (among other things).
    pub fn fetch(&self, url: &str) -> Status {
        let base = self
            .directory
            .as_path()
            .join(Self::url_to_basename(url).unwrap());

        let mut builder = self.client.get(url);
        // Don't send just yet!
        // Set some headers to see if page content has changed.

        // This break things - it destroys the file extension (if given)
        // base.set_extension("header");
        let mut header_file = base.clone();
        // Append .header to the filename.
        let filename = header_file
            .file_name()
            .unwrap()
            .to_string_lossy()
            .to_string()
            + ".header";
        header_file.set_file_name(filename);

        if header_file.exists() {
            // Ok! We have existing information.  Retrieve it.
            let old_header = load_headermap(header_file.to_str().unwrap()).unwrap();

            // Look for: ETag, Last-Modified
            if let Some(lastmod) = old_header.get("Last-Modified") {
                builder = builder.header("If-Modified-Since", lastmod);
            } else if let Some(date) = old_header.get("Date") {
                // Keep trying...
                builder = builder.header("If-Modified-Since", date);
            }

            if let Some(etag) = old_header.get("etag") {
                builder = builder.header("If-None-Match", etag);
            }
        };

        let mut result = builder.send().unwrap();

        if result.status() == 304 {
            // Cache hit!

            // println!("Cache hit! 304! {}", url);
            // base.set_extension("content");
            let fp = std::fs::File::open(base.to_str().unwrap()).unwrap();
            return Status::Cached(fp);
        }

        // Ok!  Success!
        if result.status() == 200 {
            // Success!

            // Only check content_length size, if we have been
            // given a max_size.
            if let Some(max_size) = self.max_size {
                if let Some(len) = result.content_length() {
                    if len > max_size {
                        // Is there a way to abort this safely?  Apparently yes! :D

                        // let byte = Byte::from_u64(len);
                        // let adjusted_byte = byte.get_appropriate_unit(UnitType::Binary);
                        // println!("Too Big! {adjusted_byte:.2} {}", url);
                        return Status::TooBig(len);
                    }
                }
            }

            // Only check acceptable content_types if given.
            if !self.accept.is_empty() {
                if let Some(content_type) = result.headers().get("content-type") {
                    // Check to see if accepted content.
                    let mut ct = content_type.to_str().unwrap();
                    let possible = content_type.to_str().unwrap().split_once(';');
                    if let Some((ct_part, _)) = possible {
                        ct = ct_part;
                    }
                    if !self.accept.contains(&ct.to_string()) {
                        // println!("Unacceptable content-type {} {}", ct, url);
                        return Status::Unacceptable(ct.to_string());
                    }
                }
            }

            // This seems like a lot of work, just to add something.
            // Maybe I should just change the save_header to take the url?
            /*
            // Add URL to the header, so I know what I fetched...
            let mut header2 = result.headers().clone();
            let head = reqwest::header::HeaderName::from_bytes(b"URL").unwrap();
            if let Ok(value) = reqwest::header::HeaderValue::from_str(url) {
                header2.insert(head, value);
            }
            */

            let r = save_headermap(header_file.to_str().unwrap(), url, result.headers());
            if r.is_err() {
                println!("save_headermap: {} {:?}", r.unwrap_err(), header_file);
            }

            // base.set_extension("content");
            let mut fp = std::fs::File::create(base.to_str().unwrap()).unwrap();
            let _ = result.copy_to(&mut fp);
            /*
            while let Ok(Some(chunk)) = result.chunk().await {
                let _ = fp.write(&chunk);
            }
            */

            let fp = std::fs::File::open(base.to_str().unwrap()).unwrap();
            return Status::Fetched(fp);
        } else {
            // Status error?!
            println!("Error {} {}", result.status(), url);
            return Status::ErrorStatus(u16::from(result.status()));
        }
    }
}
